/////////////////////////////////////////////////////////////////////////////
//
//    bnt.c
//    (DWL)
//
//    Big number function library. This implementation uses the Big 
// structure which defines digits a byte-wide. This is very inefficient. 
// Its main purpose in life is to emulate the big number functions 
// implemented in the DS83C950 so that I can step through these functions
// and debug a bunch of nasty 8051 code.
//    The ModExp function is a simpleton implementation that is only useful
// with small exponents (it's too slow otherwise).
//
// Version 1.00 : Initial creation
// 12/2/96
//
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <process.h>
#include <memory.h>
#include <string.h>

#include "bnt.h"

/////////////////////////////////////////////////////////////////////////////
//
//    Remove leading 0s.
//
void nlz(pBig A)
{
   while (!A->Digit[A->Len-1] && A->Len)
      --A->Len;
}

/////////////////////////////////////////////////////////////////////////////
void ShrinkIt(pBig A)
{
   // Rip out leading sign digits
   if (A->Digit[A->Len-1] == 0xFF) // Negative number
   {
      while ((A->Digit[A->Len-1] == 0xFF) && (--A->Len));

      // Don't loose our sign
      if (!(A->Digit[A->Len-1] & 0x80))
      {
         A->Digit[A->Len] = 0xFF;
         ++A->Len;
      }
   }
   else // Positive number
   {
      if (A->Len)
      {
         while ((!A->Digit[A->Len-1]) && (--A->Len));

         // Don't want to mistake number for negative
         if (A->Digit[A->Len-1] & 0x80)
         {
            A->Digit[A->Len] = 0;
            ++A->Len;
         }
      }
   }
}

/////////////////////////////////////////////////////////////////////////////
void Add(Big A, Big B, pBig C)
{
   pBig   BigAddend, LittleAddend;
   uchar  c, Sign;
   ushort w, i;

   if (A.Len < B.Len)
   {
      BigAddend    = &B;
      LittleAddend = &A;
   }
   else
   {
      BigAddend    = &A;
      LittleAddend = &B;
   }

   // Get sign of shorter number
   Sign = LittleAddend->Digit[LittleAddend->Len-1] & 0x80 ? 1 : 0;

   // Sign extend BigAddend to handle carries
   BigAddend->Digit[BigAddend->Len] = 
      (BigAddend->Digit[BigAddend->Len-1] & 0x80) ? 0xFF : 0;
   ++BigAddend->Len;

   c = 0;
   for (i = 0; i < LittleAddend->Len; i++)
   {
      w = BigAddend->Digit[i] + LittleAddend->Digit[i] + c;
      C->Digit[i] = w & 0xFF;
      c = (w > 0xFF) ? 1 : 0;
   }

   for (i = LittleAddend->Len; i < BigAddend->Len; i++)
   {
      w = BigAddend->Digit[i] + c;
      w += Sign ? 0xFF : 0;
      C->Digit[i] = w & 0xFF; 
      c = (w > 0xFF) ? 1 : 0;
   }

   C->Len = BigAddend->Len;

   if (c)
   {
      // Did we add a negative number to a smaller positive number ?
      if (C->Digit[C->Len-1])
      {
         C->Digit[C->Len] = Sign ? 0xFF : 1;
         ++C->Len;
      }
   }
   
   ShrinkIt(C);
}

/////////////////////////////////////////////////////////////////////////////
//
//   This routine is a quick divide by 2. It preserves the sign of the
// big number.
//
void Shr_1(pBig A)
{
   uchar lsb;
   ushort i;

   // Save sign of A
   uchar msb = A->Digit[A->Len-1] & 0x80;

   // Do the big shift
   for (i = A->Len-1; i != 0; i--)
   {
      lsb = A->Digit[A->Len-1-i] & 0x01;
      A->Digit[A->Len-1-i] >>= 1;
      A->Digit[A->Len-1-i] |= msb;
      msb = lsb ? 0x80 : 0;
   }
}

/////////////////////////////////////////////////////////////////////////////
//
//    Shift a big number right SNum places. The sign bit is not preserved. 
//
void ShrU(pBig A, ushort SNum)
{
   ushort ByteCnt = SNum >> 3;
   uchar  BitCnt  = SNum % 8;
   ushort i;

   if (BitCnt)
   {
      uchar Mask     = 0xFF >> (8 - BitCnt);
      uchar LastBits = 0;
      uchar NewBits;

      for (i = A->Len; i > 0; --i)
      {
         NewBits = (A->Digit[i-1] & Mask) << (8 - BitCnt);
         A->Digit[i-1] >>= BitCnt;
         A->Digit[i-1] |= LastBits;
         LastBits = NewBits;
      }
   }

   A->Len -= ByteCnt;
}

/////////////////////////////////////////////////////////////////////////////
//
//   This routine is a quick multiply by 2. It preserves the sign of the
// big number.
//
void Shl_1(pBig A)
{
   uchar  msb, lsb = 0;
   ushort i;

   // Save sign of A
   uchar s = A->Digit[A->Len-1] & 0x80;

   // Do the big shift
   for (i = 0; i < A->Len; i++)
   {
      msb = A->Digit[i] & 0x80;
      A->Digit[i] <<= 1;
      A->Digit[i] |= lsb;
      lsb = (msb == 0x80) ? 1 : 0;
   }

   A->Digit[A->Len-1] |= s;
}

/////////////////////////////////////////////////////////////////////////////
//
//    This shift routine allows a number to grow as opposed to falling off
// into the bit bucket.  
//
void ShlU(pBig A, ushort SNum)
{
   ushort ByteCnt  = SNum >> 3;
   uchar  BitCnt   = SNum % 8;
   ushort i;
   uchar  Mask     = 0xFF << (8 - BitCnt);
   uchar  LastBits = 0;
   uchar  NewBits;

   if (BitCnt)
   {
      A->Digit[A->Len++] = 0;

      for (i = 0; i < A->Len; i++)
      {
         NewBits = (A->Digit[i] & Mask) >> (8 - BitCnt);
         A->Digit[i] <<= BitCnt;
         A->Digit[i] |= LastBits;
         LastBits = NewBits;
      }
   }

   if (ByteCnt)
   {  
      uchar TmpBuff[MAX_DIG];
     
      memcpy(TmpBuff, A->Digit, A->Len);
      memcpy(&A->Digit[ByteCnt], TmpBuff, A->Len);
      A->Len += ByteCnt;
      memset(A->Digit, 0x00, ByteCnt);
   }
}

/////////////////////////////////////////////////////////////////////////////
void Negate(pBig A)
{
   ushort i;

   // 1's complement digits
   for (i = 0; i < A->Len; i++)
      A->Digit[i] ^= 0xFF;
   // Add 1
   i = 0;
   while (!(++A->Digit[i]) && (++i < A->Len));
}

/////////////////////////////////////////////////////////////////////////////
void Sub(Big A, Big B, pBig C)
{
   Big Tmp;

   Assign(&Tmp, &B);
   Negate(&Tmp);
   // C = A + (-B)
   Add(A, Tmp, C);
}

/////////////////////////////////////////////////////////////////////////////
void Mul(Big A, Big B, pBig C)
{
   Big    Product;
   uchar  x;
   ushort i, j, w;

   if ((!A.Len) || (!B.Len))
   {
      C->Len = 0;
      return;
   }

   Product.Len = A.Len + B.Len;
   memset(Product.Digit, 0, MAX_DIG);

   for (i = 0; i < B.Len; i++)
   {
      x = 0;
      for (j = 0; j < A.Len; j++)
      {
         w = A.Digit[j] * B.Digit[i] + Product.Digit[i+j] + x;
         Product.Digit[i+j] = w & 0xFF;
         x = (w & 0xFF00) >> 8;
      }

      Product.Digit[i+A.Len] = x;
   }

   while (Product.Len && (!Product.Digit[Product.Len-1]))
      --Product.Len;

   Assign(C, &Product);
}

/////////////////////////////////////////////////////////////////////////////
void Div(Big Dividend, Big Divisor, pBig Quotient, pBig Remainder)
{
   ushort i, j, k;
   ushort LenU, LenV;
   ulong  v1, v2, u0, u1, u2;
   ulong  w;
   uchar  d, QHat;
   int    cr;
   Big    Tmp;
   Bool   Borrow;

   Quotient->Len  = 0;
   Remainder->Len = 0;

   // No leading 0s
   nlz(&Dividend); nlz(&Divisor);

   switch (Divisor.Len)
   {
      // Divide by 0, this probably isn't too good
      case 0:
         printf("Divide by 0 error !!!\n");
         exit(1);
      break;

      // Short division, yipee !!!
      case 1:
         // v1 is the only digit in the divisor
         v1 = Divisor.Digit[0];
         u0 = 0;

         for (i = Dividend.Len; i > 0; i--)
         {
            w = (u0 * 256) + Dividend.Digit[i-1];
            Quotient->Digit[i-1] = w / v1;
            u0 = w % v1;
            ++Quotient->Len;
         }

         // Set remainder
         AssignByte(Remainder, u0 & 0xFF);
      break;

      // Long division, DOH !!!
      default:
         LenU = Dividend.Len;
         LenV = Divisor.Len;

         cr = CompUnsigned(Dividend, Divisor);

         if (cr == 0)
         {
            AssignByte(Quotient, 1);
            AssignByte(Remainder, 0);
         }
         else if (cr == -1)
         {
            Assign(Remainder, &Dividend);
            AssignByte(Quotient, 0);
         }
         else
         {
            // Compute multiplicative constant (Knuth D1)
            d = 256 / (Divisor.Digit[Divisor.Len-1] + 1);
        
            if (d > 1)
            {
               AssignByte(&Tmp, d);
               Mul(Divisor, Tmp, &Divisor);
               Mul(Dividend, Tmp, &Dividend);
            }

            // Get 2 MSDs of divisor
            v1 = Divisor.Digit[Divisor.Len-1];
            v2 = Divisor.Digit[Divisor.Len-2];

            ++LenU; 
            if (LenU > Dividend.Len)
               Dividend.Digit[Dividend.Len++] = 0;
 
            for (j = LenU - 1; j >= Divisor.Len; j--)
            {
               // Get 3 MSDs of dividend
               u0 =  j       ? Dividend.Digit[j]   : 0;
               u1 = (j >= 1) ? Dividend.Digit[j-1] : 0;
               u2 = (j >= 2) ? Dividend.Digit[j-2] : 0;

               // Compute new QHat
               QHat = (v1 == u0) ? 0xFF : (u0*256 + u1) / v1;

               // Knuth's quick test.  Try to make QHat small enough
               while ((v2*QHat) > ((u0*256 + u1 - QHat*v1)*256 + u2))
                  --QHat;

               AssignByte(&Tmp, QHat);
               Mul(Tmp, Divisor, &Tmp);

               // Pad subtrahend with leading 0s
               while (Tmp.Len <= LenV)
                  Tmp.Digit[Tmp.Len++] = 0;

               Borrow = FALSE;

               for (i = 0; i < Tmp.Len; i++)
               {
                  if (Borrow)
                  {
                     ++Tmp.Digit[i];

                     k = i;
                     while (Borrow && (!Tmp.Digit[k++]) && (k < Tmp.Len))
                        ++Tmp.Digit[k];
                  }

                  Borrow = (Dividend.Digit[j-LenV+i] < Tmp.Digit[i]);
                  Dividend.Digit[j-LenV+i] = 
                     (Dividend.Digit[j-LenV+i] - Tmp.Digit[i]) & 0xFF;
               }
            
               // Subtraction was negative, add back 1 divisor
               if (Borrow)
               {
                  --QHat;
                  Tmp.Digit[Tmp.Len++] = 0;
                  w = 0;
                  for (i = 0; i < Tmp.Len; i++)
                  {
                     w >>= 8;
                     w = w + Dividend.Digit[j-LenV+i] + Divisor.Digit[i];
                     Dividend.Digit[j-LenV+i] = w & 0xFF;
                  }
               }
            
               // All that for 1 lousy byte of quotient
               Quotient->Digit[j-LenV] = QHat;
            }

            Quotient->Len = LenU - LenV;
            nlz (Quotient);

            w = 0;
            for (i = LenV; i > 0; i--)
            {
               w = (w << 8) + Dividend.Digit[i-1];
               Dividend.Digit[i-1] = (w / d) & 0xFF;
               w = w - Dividend.Digit[i-1] * d;
            }

            Dividend.Len = LenV;
            nlz(&Dividend);

            Assign(Remainder, &Dividend);
         }
      break;
   }
}

/////////////////////////////////////////////////////////////////////////////
//
//    Returns the number of bits in A.
//
ushort BitCount(pBig A)
{
   uchar  x;
   ushort BitCount = 8;

   // Remove leading 0s.
   nlz(A);

   if (A->Len)
   {
      x = A->Digit[A->Len-1];

      while (!(x & 0x80))
      {
         --BitCount;
         x <<= 1;
      }
   }

   return A->Len ? ((A->Len-1) << 3) + BitCount : 0;
}

/////////////////////////////////////////////////////////////////////////////
//
//    Compute: R = A Mod B.
//
void SimpleReduction(Big T, Big N, pBig R)
{
   ushort i, k;

   // Rip out any leading 0s.
   nlz(&N); nlz(&T);
 
   k = BitCount(&N);

   // Make sure T is small enough
   if (BitCount(&T) < (k << 1) + 1)
   {
      Assign(R, &T);
      ShlU(&N, k - 1);
      
      for (i = 0; i < k; i++)
      {
         if (CompUnsigned(*R, N) != -1)
         {
            // Pad with 1 0 for subtraction of positive numbers
            N.Digit[N.Len++] = R->Digit[R->Len++] = 0;
            Sub(*R, N, R);
            nlz(&N); nlz(R);
         }

         ShrU(&N, 1);
      }
   }
}

/////////////////////////////////////////////////////////////////////////////
//
//    Compute C = M^E Mod N.
//
void ModExp(Big M, Big E, Big N, pBig C)
{
   ushort i, j;
   Big    Q;

   // Rip out leading 0s.
   nlz(&M); nlz(&E); nlz(&N); 

   AssignByte(C, 1);
   for (i = 0; i < E.Len; i++)
   {
      uchar ExpByte = E.Digit[i];
      for (j = 0; j < 8; j++)
      {
         if (ExpByte & 1)
         {
            Mul(*C, M, C);
            //SimpleReduction(*C, N, C);
            Div(*C, N, &Q, C);
         }

         ExpByte >>= 1;
         Mul(M, M, &M);
         //SimpleReduction(M, N, &M);
         Div(M, N, &Q, &M);
      }
   }
}

/////////////////////////////////////////////////////////////////////////////
void DumpBig(pBig x)
{
   Big    tmp;
   uchar  s;
   ushort i;
  
   Assign(&tmp, x);

   s =(tmp.Digit[tmp.Len-1] & 0x80) ? 1 : 0;
      
   if (s)
   {
      Negate(&tmp);
      printf("-");
   }

   if (!x->Len)
   {
      printf("0");
   }
   else
   {
      for (i = 0; i < tmp.Len; i++)
         printf("%2.2X", tmp.Digit[tmp.Len-i-1]);
   }

   printf("H\n");
}

/////////////////////////////////////////////////////////////////////////////
void DumpUnsigned(pBig x)
{
   ushort i;

   if (!x->Len)
   {
      printf("0");
   }
   else
   {
      for (i = 0; i < x->Len; i++)
         printf("%2.2X", x->Digit[x->Len-i-1]);
   }
   
   printf("H\n");
}

Big  b, d;
Big  m, n;
Bool swapped;

/////////////////////////////////////////////////////////////////////////////
void swap(pBig A, pBig B)
{
   Big tmp;

   tmp.Len = A->Len;
   memcpy(tmp.Digit, A->Digit, A->Len);
   A->Len = B->Len;
   memcpy(A->Digit, B->Digit, A->Len);
   B->Len = tmp.Len;
   memcpy(B->Digit, tmp.Digit, tmp.Len);
}

/////////////////////////////////////////////////////////////////////////////
void reduce(pBig A, pBig B)
{
   while (!(A->Digit[0] & 1))
   {
      Shr_1(A);

      if (B->Digit[0] & 1)
      {
         if (!(B->Digit[B->Len-1] & 0x80))
            Sub(*B, m, B);
         else
            Add(*B, m, B);
      }

      Shr_1(B);
   }
}

/////////////////////////////////////////////////////////////////////////////
//
//    Returns 0 if A = B. If A > B Comp returns 1, otherwise comp 
// returns -1.  Really dirty implementation.
//
int Comp(pBig A, pBig B)
{
   Big Diff;

   // A is positive and B is negative
   if (!(A->Digit[A->Len-1] & 0x80) && (B->Digit[B->Len-1] & 0x80))
      return 1;

   // B is positive and A is negative
   if (!(B->Digit[B->Len-1] & 0x80) && (A->Digit[A->Len-1] & 0x80))
      return -1;

   Sub(*A, *B, &Diff);

   // A = B if Diff is 0
   if (!Diff.Len || ((Diff.Len == 1) && (!Diff.Digit[0])))
      return 0;

   // A and B negative
   if ((B->Digit[B->Len-1] & 0x80) && (A->Digit[A->Len-1] & 0x80))
      return (Diff.Digit[Diff.Len-1] & 0x80) ? 1 : -1;

   // A and B are both positive   
   return (Diff.Digit[Diff.Len-1] & 0x80) ? -1 : 1;
}

/////////////////////////////////////////////////////////////////////////////
//
//    Returns 0 if A = B. If A > B Comp returns 1, otherwise comp 
// returns -1. 
//
int CompUnsigned(Big A, Big B)
{
   int Comp;

   nlz(&A); nlz(&B);
  
   if (A.Len == B.Len)
   {
      ushort i = A.Len;
      while ((A.Digit[i-1] == B.Digit[i-1]) && --i);
      if (i)
      {
         Comp = A.Digit[i-1] > B.Digit[i-1] ? 1 : -1;
      }
      else
      {
         Comp = 0;
      }
   }
   else
   {
      Comp = (A.Len > B.Len) ? 1 : -1;
   }

   return Comp;
}

/////////////////////////////////////////////////////////////////////////////
//
//   A = B.
//
void Assign(pBig A, pBig B)
{
   A->Len = B->Len;
   memcpy(A->Digit, B->Digit, B->Len);
}

/////////////////////////////////////////////////////////////////////////////
void AssignByte(pBig A, uchar x)
{
   if (!x)
   {
      A->Len = 0;
   }
   else
   {
      A->Len = 1;
      A->Digit[0] = x;
   }
}

/////////////////////////////////////////////////////////////////////////////
void ModInv(pBig m, pBig n)
{
   Big  u2, u3;
   Big  v2, v3;
   uchar f; 

   f = 0;

   // Compute powers of 2 in gcd(m, n)
   while (!(m->Digit[0] & 1) && !(n->Digit[0] & 1))
   {
      Shl_1(m); Shl_1(n); ++f;
   }

   swapped = FALSE; 
   if (!(m->Digit[0] & 1))
   {
      swapped = TRUE;
      swap(m, n);
   }

   // Initialize u vector
   u2.Len = 0;
   u2.Digit[0] = 0;
   Assign(&u3, m);
   
   // Initialize v vector
   v2.Len = 1;
   v2.Digit[0] = 1;
   Assign(&v3, n);

   if (!(v3.Digit[0] & 1))
      reduce(&v3, &v2);
   
   if (Comp(&u3, &v3) == -1) // If (u3 < v3)
   {
      swap(&u2, &v2); 
      swap(&u3, &v3); 
   }

   Sub(u2, v2, &u2);
   Sub(u3, v3, &u3);
   
   while (!((u3.Len == 0) || ((u3.Len == 1) && (!u3.Digit[0]))))
   {
      reduce(&u3, &u2);

      if ((u3.Len == 1) && (u3.Digit[0] == 1))
         break; 

      if (Comp(&u3, &v3) == -1) // If (u3 < v3)
      {
         swap(&u2, &v2); 
         swap(&u3, &v3); 
      }

      Sub(u2, v2, &u2);
      Sub(u3, v3, &u3); 
   }

   if ((u3.Len == 0) || ((u3.Len == 1) && (!u3.Digit[0])))
   {
      Assign(&b, &v2);
      Assign(&d, &v3);
      // !!!!!!! Fix this !!!!!!!!
      //Shl(&d, f);
   }
   else
   {
      Assign(&b, &u2);
      d.Len      = 1;
      d.Digit[0] = 1;
      // !!!!!!! Fix this !!!!!!!!
      // Shl(&d, f);
   }
}

Big One;

/////////////////////////////////////////////////////////////////////////////
//
//    Test the suspected prime using n randomly generated witnesses.
//
Bool TestPrime(Big sp, ushort n)
{
   Big    w, e, d;
   ushort i, j;

   One.Len = One.Digit[0] = 1;

   Sub(sp, One, &e);
   
   i = 0;
   while (i < n)
   {

      // Randomly generate a witness
      w.Len = sp.Len-1;
      for (j = 0; j < w.Len; j++)
         w.Digit[j] = rand() & 0xFF;
      
      ModExp(w, e, sp, &d);

      if (CompUnsigned(d, One) != 0)
         return FALSE;

      ++i;
   }

   return TRUE;
}
